from models import FmuModel
from models import DymolaModel
from models import Estim8Model

from estim8 import Estimator
from estim8 import ParameterMapping
from estim8 import LossCalculator
from estim8 import Workers
from estim8 import ErrorModel

from visualization import Visualization

from utils import Utils



from utils import Utils
import os
import sys
import sdf
from warnings import warn
import abc
import uuid
import fmpy
import time
from fmpy.fmi2 import FMU2Slave


class Estim8Model(object):

    def __init__(self, name:str, model_type:str):
        self.name               = name
        self._model_type        = model_type
        self.model_path         = None
        self._parameters        = None
        self._observations      = None
        self._variables         = None
        self._loaded            = False



    @abc.abstractmethod
    def initialize(self, for_est=False):
        """
        Defines method for steps of initialization of models
        """

    @abc.abstractmethod
    def simulate(self, t0:float, t_end:float, stepsize:float, observe:list=None, parameters:dict=None,
                 tolerance:float=1e-4, res_file=None, for_est=None):
        """
        Defines simulation method
        """

    # Defining Properties:
    @property
    def variables(self) -> dict:
        return self._variables

    @variables.setter
    def variables(self, values: dict):
        if values is not None and not isinstance(values, dict):
            raise TypeError('variables must be either None or dictionary')
        self._variables = values


    @property
    def parameters(self) -> dict:
        return self._parameters

    @parameters.setter
    def parameters(self, values: dict):
        if values is not None and not isinstance(values, dict):
            raise TypeError('parameters must be either None or dictionary')
        if self._variables is not None:
            klv = [str(item) for item in values.keys()]
            test = [str(item) for item in self._variables.keys()]
            _missing = list(set(klv) - set(test))
            if _missing:
                msg = "One or more keys specified for parameters is not contained in the model:\n{}".format(
                    _missing)
                warn(msg)
        self._parameters = values


    @property
    def observations(self) -> list:
        return self._observations

    @observations.setter
    def observations(self, values: list):
        if values is not None and not isinstance(values, list):
            raise TypeError('observations must be either None or list')
        if self._variables is not None:
            klv = [str(item) for item in values]
            test = [str(item) for item in self._variables.keys()]
            _missing = list(set(klv) - set(test))
            if _missing:
                msg = "One or more keys specified for observation is not contained in the model:\n{}".format(
                    _missing)
                warn(msg)
        self._observations = values



class DymolaModel(Estim8Model):
    """
    DymolaModel is the class connecting Python and Dymola. By creating an instance of DymolaModel, all required paths
    are set automatically. Its methods are used for simulating in Dymola.
    """

    def __init__(self, name:str, model_name:str=None, model_path:str=None, dymola_path:str=None):
        """
        Arguments
        ---------
            name : str
                Name of the model (also referred to as "problem") without ".mo" ending. Constructed by super class

        Keyword arguments
        -----------------
            model_name : str
                same as name. Initialized for further use...
            model_path : str
                Path on your computer to the Dymola file. Generated automatically. If not in same or in parent directory,
                specifying it can improve speed
            dymola_path : str
                Path to "dymola.egg" file. Specify if not automatically found.py
        """
        super().__init__(name, model_type='dymola')
        self.dymola_path        = Utils.find_dymola_egg(path=dymola_path)
        if model_name is None:
            self.model_path     = Utils.find_model_path(name, model_path, self._model_type)
            self.model_name     = name
        else:
            self.model_path     = Utils.find_model_path(model_name, model_path)
            self.model_name     = model_name

        #Later on used attributes
        self._dymola            = None

        # Adding path to dymola egg
        sys.path.insert(0, self.dymola_path)



    def info(self):
        print(f'Model name:\t\t {self.model_name}')
        print(f'Model path:\t\t {self.model_path}')
        print(f'Dymola location:\t {self.dymola_path}')
        print(f'Model loaded:\t\t {self._loaded}')
        print(f'Model variables:\n{self.variables.keys()}')
        print(f'Model parameters:\n{self.parameters.keys()}')



    def load_model(self, for_est=False):
        """
        Initializes Dymola model such that it is executable by other methods. Creates an instance of DymolaInterface as
        self._dymola.

        Warns
        -----------------
            Dymola Exception
                Exceptions raised by DymolaInterface
            User warning
                Model already loaded
        """

        # Import Dymola Interface
        from dymola.dymola_interface import DymolaInterface
        from dymola.dymola_exception import DymolaException

        if not self._loaded:
            # Initialize Interface object
            self._dymola = None

            try:
                # Open Interface
                self._dymola = DymolaInterface()

                # Open Model
                ok = self._dymola.openModel(path=self.model_path, mustRead=False, changeDirectory=True)
                if ok:
                    if not for_est:
                        print(f'Opened Model {self.model_name}')
                else:
                    print('Model could not be opened')
                    print(self._dymola.getLastErrorLog())

                # Translate Model
                ok = self._dymola.translateModel(self.name)
                if ok:
                    if not for_est:
                        print('Translation successful')
                        print(f'Ready to use {self.model_name}')
                    self._loaded = True
                else:
                    print('Translation failed')
                    print(self._dymola.getLastErrorLog())

            except DymolaException as ex:
                warn(('Error: ' + str(ex)))

        else:
            warn(f'Model {self.model_name} is already loaded')



    def initialize(self, load:bool=True, read_vars:bool=True, close_after:bool=True, for_est=False):
        """
        Executes all steps required for running a Dymola model.

        Keyword Arguments:
        ------------------
            load        :   boolean
                Executing load and translation step if True
            read_vars   :   boolean
                Executing retrieve variables step if True
            close_after:   boolean
                Closing dymola model after initialization if True
        """

        # Loading step
        __v = []
        if load:
            self.load_model(for_est=for_est)
            __ok = True
        elif not load and self._loaded:
            __ok = True
        else:
            __ok = False

        # Variable retrieving step
        if read_vars and __ok:
            self.retrieve_variables()
            __ok = True
        else:
            __ok = False

        if close_after and self._loaded:
            self.close_model()

        return True


    def close_model(self):
        """
        Closes an opened dymola model.

        Raises:
            RuntimeError:
                Model not loaded
        """
        if not self._loaded:
            raise RuntimeError('The model is not loaded')
        self._dymola.close()
        self._dymola = None
        self._loaded = False




    def kill_dymola_processes(self):
        """
        Method for killing all DymolaYYYYx.exe processes. Reads your dymola version number to formulate a kill command
        and passes it to os.system().
        """
        warn('This command will kill all Dymola related processes.')
        _input = input('Type \"No\" for stopping')
        if (_input=='No') or (_input=='no'):
            pass
        else:
            self._loaded = False
            return exec('os.system("taskkill /F /im Dymola.exe")')



    def retrieve_variables(self, sort=True, output:bool=False):
        """
        Uses the get_variable_list function to retrieve the variable names and additionally creates a dict with values 0
        and also changes the property \"variables\" of the model. For more information type:
            help(DymolaModel.get_variable_list)

        Keyword Arguments:
        ------------------
            sort    :   bool
                Sorts variables into parameters and observations if True
            output  :   bool
                Print output if True
        """

        # Load model if required
        if self._loaded:
            pass
        else:
            self.load_model()

        # Random result file name
        rn = str(uuid.uuid4())

        # Execute Pseudo Simulation
        _sim = self._dymola.simulateExtendedModel(problem=self.model_name,
                                                  startTime=0.0,
                                                  stopTime=1.0,
                                                  outputInterval=1,
                                                  method="Dassl",
                                                  tolerance=0.0001,
                                                  autoLoad=True,
                                                  resultFile=rn
                                                  )

        # Load simulated data and delete the file
        sdf_data    = sdf.load(rn+".mat")
        os.remove(rn+".mat")

        # Read variable names
        _vars       = [p.name for p in sdf_data.datasets]

        # Remove Time-Attribute (not needed)
        if "Time" in _vars:
            _vars.pop(0)

        # Initialize dicts
        _v = {}
        _p = {}
        _o = []

        # Check if sort is activated
        if sort:

            for _var_name in _vars:
                # If data is an array -> State Variable
                if not isinstance(sdf_data[_var_name].data, float) and not isinstance(sdf_data[_var_name].data, int):
                    _o.append(_var_name)
                    _v[_var_name] = 1.0
                # If data is scalar -> Parameter
                else:
                    _v[_var_name] = 1.0
                    _p[_var_name] = sdf_data[_var_name].data

            # Assign to Attributes
            self.variables      = _v
            self.observations   = _o
            self.parameters         = _p
        else:
            _v = {}
            for _i in _vars:
                _v[_i] = 1.0

            # Assign to Attributes
            self.variables = _v

        if output and sort:
            return _v, _p, _o
        elif output and not sort:
            return _v



    def simulate(self, t0:float, t_end:float, stepsize:float, observe:list=None, parameters:dict=None,
                 tolerance:float=1e-4, res_file:str='dsres', for_est=False):
        """
        Simulates the model given the specified settings.

        Arguments:
        ----------
            t0          :   float
                Start time
            t_end       :   float
                Stop time
            stepsize    :   float
                Time interval increment (linear steps)

        Keyword Arguments:
        ------------------
            observe     :   list[str]
                Observations to retrieve from simulation (Default: Model specification)
            parameters  :   dict
                Parameters used for the simulation. (Default: Model specification)
            tolerance   :   float
                Integration tolerance for dymola model (Default: 1e-4)
            res_file    :   str
                Name of the result file given without \".mat\" ending. Input required for multiprocessing to avoid
                overwriting files of other threads.

        Raises:
        -------
            UserWarning :
                Model not loaded or not valid
        """

        # Use default if not further specified
        if observe is None:
            observe         = self.observations
        if parameters is None:
            par_names      = [str(item) for item in self.parameters.keys()]
            par_vals       = [item for item in self.parameters.values()]
        else:
            par_names = [str(item) for item in parameters.keys()]
            par_vals = [item for item in parameters.values()]


        # Check if model is loaded
        if not self._loaded:
            if not for_est:
                warn(f'Model was not loaded. Loading {self.model_name}...')
            self.initialize(close_after=False, for_est=for_est)
            _loaded_here = True
        else:
            _loaded_here = False

        # Validate model
        if self._dymola.checkModel(self.model_name, simulate=True):
            pass
        else:
            warn('Model is not valid. Check implementation and Dymola installation')

        try:
            self._dymola.simulateExtendedModel(
                                            problem         = self.model_name,
                                            startTime       = t0,
                                            stopTime        = t_end,
                                            outputInterval  = stepsize,
                                            initialNames    = par_names,
                                            initialValues   = par_vals,
                                            finalNames      = observe,
                                            tolerance       = tolerance,
                                            resultFile      = res_file,
                                           )
        except:
            raise ChildProcessError(f'{self._dymola.getLastErrorLog()}')

        if _loaded_here and not for_est:
            self.close_model()
            print('Model has been closed again. To load it use \".load_model()\".')

        sdf_data = sdf.load(res_file + '.mat')
        _sim = {}
        _par = {}
        for obs in sdf_data.datasets:
            if obs.name in self.parameters.keys():
                _par[obs.name] = obs.data
            else:
                _sim[obs.name] = obs.data


        return [_sim, _par]


    def sim_parallel(self, t0:float, t_end:float, stepsize:float, observe:list=None, parameters:dict=None,
                 tolerance:float=1e-4, for_est=True, in_folder:str=None):

        # Use default if not further specified
        if observe is None:
            observe = self.observations
        if parameters is None:
            par_names = [str(item) for item in self.parameters.keys()]
            par_vals = [item for item in self.parameters.values()]
        else:
            par_names = [str(item) for item in parameters.keys()]
            par_vals = [item for item in parameters.values()]

        # Change folder
        if in_folder is not None:
            os.chdir(in_folder)

        _sim = None
        _par = None

        # Load model for estimation
        if not self._loaded:
            self.load_model(for_est=for_est)

        # Validate model
        if self._dymola.checkModel(self.model_name, simulate=True):
            pass
        else:
            warn('Model is not valid. Check implementation and Dymola installation')

        # Generate random filename
        res_file = 'Temp_Sim_' + str(uuid.uuid4())

        try:
            self._dymola.simulateExtendedModel(
                problem=self.model_name,
                startTime=t0,
                stopTime=t_end,
                outputInterval=stepsize,
                initialNames=par_names,
                initialValues=par_vals,
                finalNames=observe,
                tolerance=tolerance,
                resultFile=res_file,
            )
        except:
            raise ChildProcessError(f'{self._dymola.getLastErrorLog()}')


        try:
            sdf_data = sdf.load(res_file + '.mat')
            _sim = {}
            _par = {}
            for obs in sdf_data.datasets:
                if obs.name in self.parameters.keys():
                    _par[obs.name] = obs.data
                else:
                    _sim[obs.name] = obs.data

            try:
                os.remove(res_file + '.mat')
            except FileNotFoundError:
                pass
            except PermissionError:
                pass

        except FileNotFoundError:
            _sim = None
            _par = None


        if _sim is None:
            warn('returning None')

        return [_sim, _par]


class FmuModel(Estim8Model):
    
    def __init__(self, name:str, model_path:str=None):
        super().__init__(name, model_type='fmu')

        self.model_name  = name + '.fmu'
        self.model_path  = Utils.find_model_path(name, model_path=model_path, model_type=self._model_type)
        self.model_desc  = fmpy.read_model_description(self.model_path)
        self._unzipdir   = fmpy.extract(self.model_path)
        self.model       = fmpy.instantiate_fmu(self._unzipdir, self.model_desc)
        self._loaded     = True

        # Get variable reference numbers:
        self.vrs = {}
        for variable in self.model_desc.modelVariables:
            self.vrs[variable.name] = variable.valueReference

        # Setup slave
        self.fmu         = fmpy.fmi2.FMU2Slave( guid=self.model_desc.guid,
                                                modelIdentifier=self.model_desc.coSimulation.modelIdentifier,
                                                unzipDirectory=self._unzipdir,
                                                )

        # Set default FMI type for simulation
        self._fmi_type   = 'ModelExchange'


    @property
    def fmi_type(self):
        return self._fmi_type

    @fmi_type.setter
    def fmi_type(self,fmitype:str):
        if fmitype == 'ModelExchange':
            self._fmi_type = fmitype
        elif fmitype in ['CoSimulation', 'Co-Simulation']:
            self._fmi_type = None
        else:
            raise ValueError(f'{fmitype} is not a valid FMI type. Choose either "CoSimulation" or "ModelExchange".')






    def retrieve_variables(self):
        """
        retrieve_variables retrieves all variable and parameter names from model description and saving the varsiables
        as variables attribute. Subsequently it executes a pseudo simulation, to check if the values change over time to
        retireve tha values for the parameters.

        External Functions:
        -------------------
            fmpy.simulate_fmu : Simulation of FMU models
            Utils.FMU_sim2dict: Converting simulation results into dictionaries.
        """

        # Read all variable names
        var_names   = [var.name for var in self.model_desc.modelVariables]
        par_names   = [par.name for par in self.model_desc.modelVariables if par.causality == 'parameter']

        # Make pseudo simulation to retrieve values
        pseudo_sim  = fmpy.simulate_fmu(self._unzipdir,
                                        start_time      = 0,
                                        stop_time       = 10,
                                        output_interval = 2,
                                        fmu_instance    = self.model,
                                        output          = var_names,
                                        )

        # Convert results into dict
        ps_sim_dict = Utils.FMU_sim2dict(var_names, pseudo_sim)

        # Sort into parameters and observations
        pars = {}
        var  = {}
        obs  = []
        for key in ps_sim_dict.keys():
            if not key.startswith('_'):                                     # Exclude private variables
                if key in par_names:
                    pars[key]   = ps_sim_dict[key][0]
                    var[key]    = ps_sim_dict[key][0]
                elif key != 'Time':
                    obs.append(key)
                    var[key]    = 0.0


        # Add to attributes
        self.variables      = var
        self.parameters     = pars
        self.observations   = obs


    def initialize(self, fmi_type='ModelExchange', for_est=False):
        """
        Starts the retrieve_variables() method.
        """
        self.retrieve_variables()
        self.model.reset()
        self.fmi_type = fmi_type
        self._loaded = True
        if not for_est:
            print(f">> {self.name} successfully initialized")


    def simulate(self, t0:float, t_end:float, stepsize:float, observe:list=None, parameters:dict=None,
                 tolerance:float=1e-4, res_file=None, for_est=None, solver:str=None):

        # 0. Check input
        ## Model loaded?
        if not self._loaded:
            warn('Model was not initialized')
            self.initialize()
        ## New parameters given?
        if parameters is None:
            parameters          = self.parameters
        else:
            _p_temp             = self.parameters       # Temporary store model parameters
            self.parameters     = parameters            # Debug by property call
            self.parameters     = _p_temp               # Put parameters back to normal
        ## Observe only special variables?
        if observe is None:
            observe             = self.observations
        else:
            _o_temp             = self.observations
            self.observations   = observe
            self.observations   = _o_temp
        ## Is a solver specified?
        if solver is not None:
            if solver not in ['Euler','CVode']:
                raise NotImplementedError(f'Solver {solver} not available. Use "Euler" or "CVode".')
            if self.fmi_type != 'ModelExchange':
                warn(f'Solver argument ignored because it is only available for FMI type model exchange.')

        # 2. Simulate
        if self.fmi_type == 'ModelExchange':
            sim_raw = fmpy.simulate_fmu(self._unzipdir,
                                        start_time      = t0,
                                        stop_time       = t_end,
                                        output_interval = stepsize,
                                        validate        =True,
                                        output          = observe,
                                        start_values    = parameters,
                                        relative_tolerance = tolerance,
                                        initialize      = True,
                                        fmi_type        = self.fmi_type,
                                        solver          = solver,
                                        )
        else:
            sim_raw = fmpy.simulate_fmu(self._unzipdir,
                                        start_time=t0,
                                        stop_time=t_end,
                                        output_interval=stepsize,
                                        validate=True,
                                        fmu_instance=self.model,
                                        output=observe,
                                        start_values=parameters,
                                        )

        # 3. Convert simulation results
        sim = Utils.FMU_sim2dict(observe, sim_raw)

        # 4. Reset model
        self.model.reset()

        return sim, parameters


import numpy as np
import pandas.core.frame

from utils import Utils
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from warnings import warn
from estim8 import Estimator
import pandas as pd




class Visualization:

    @staticmethod
    def plot_sim(sim_data, observe: list = None):
        """
        Function for plotting time trajectories of simulation data. Either plotting all observations contained in
        sim_data, or only those specified in observe.

        Arguments:
        ----------
            sim_data    :   dict / list[dict]
                Data to be plotted.

        Keyword Arguments:
        ------------------
            observe     :   list[str]
                List of keys from sim_data which must be plotted.

        Raises:
        -------
            TypeError:
                sim_data neither dict nor list[dict]
            TypeError:
                observe entries are not string
            UserWarning:
                Some keys specified in observe are not in sim_data.
        """

        # Check sim_data
        if isinstance(sim_data, list):
            sim_data = sim_data[0]
        if not isinstance(sim_data, dict):
            raise TypeError(f'sim_data needs to be of type dict or list. In case of list, the first element must contain a dict. Provided data was {type(sim_data)}')

        # Check observations to plot
        if observe is None:
            obs_list = [str(item) for item in sim_data.keys() if item != "Time"]
        else:
            obs_list = observe

        # Check observe
        if not Utils.check_dtype(obs_list,str):
            raise TypeError('Entries of observe are not string type')
        if not all(items in sim_data.keys() for items in obs_list):
            _blacklist = list(set(obs_list)-set(sim_data.keys()))
            warn(f'The observations {_blacklist} are not contained in sim_data. Continuing without...')
            obs_list = list(set(obs_list)-set(_blacklist))

        # Define subplot grid:
        _v_size = int(len(obs_list) / 2) + len(obs_list) % 2
        _h_size = 2

        _fig, _axs = plt.subplots(_v_size, _h_size)

        # Set figure size to jupyter notebook range
        _fig.set_size_inches(18, 4.5 * _v_size)

        # Excluding empty plot
        if (len(obs_list) % 2 != 0):
            try:
                _fig.delaxes(_axs[_v_size - 1, 1])
            except IndexError:
                _fig.delaxes(_axs[1])


        # Initialize subplot enumerators
        _h = 0
        _v = 0

        # Cycle colors
        colors = plt.rcParams["axes.prop_cycle"]()

        # Create plots
        for obs in obs_list:
            c = next(colors)["color"]  # Select next color
            if (_h_size == 1) and (_v_size == 1):
                _axs.plot(sim_data["Time"], sim_data[obs], color=c)
                _axs.set_xlabel("Time")
                _axs.set_ylabel(str(obs))
                _h += 1
            elif (_h_size == 1) or (_v_size == 1):
                _axs[_h].plot(sim_data["Time"], sim_data[obs], color=c)
                _axs[_h].set_xlabel("Time")
                _axs[_h].set_ylabel(str(obs))
                _h += 1
            else:
                _axs[_v, _h].plot(sim_data["Time"], sim_data[obs], color=c)
                _axs[_v, _h].set_xlabel("Time")
                _axs[_v, _h].set_ylabel(str(obs))
                if _h == _h_size - 1:
                    _v, _h = _v + 1, 0  # if _h is at position 1, set zero and increase _v
                else:
                    _h += 1  # else increase _h

        plt.show()


    @staticmethod
    def plot_estimates(estimates:dict=None, estimator_object:Estimator=None, data:dict=None, only_measured:bool=False):
        """
        Plotting estimates and data together. Uses data of estimator_object if argument data is not specified.

        Keyword Arguments:
        ------------------
            estimates           :   dict
                Dictionary containing the estimated parameters. Non-specified parameters are interpreted as replicate or
                model default values.
            estimator_object    :   Estimator
                Object of estimator class (should be the Estimator that carried out the estimation). Used for simulation,
                parameter_mapping and providing data if not specified.
            data                :   dict
                Dataset to compare with. Using data specified in Estimator object if not given.
            only_measured       :   bool
                Plots only the simlation data for which measurements exist

        External Functions:
        -------------------
            estim8.Estimator.data (setter method):
                For checking data input

        Raises:
        -------
            IOError:
                No Input given for necessary keywords
            TypeError:
                Wrong datatypes from input
            ValueError:
                Wrong input structure / wrong dtypes in dicts
        """


        """ Input Processing """

        # Check estimate argument
        if estimates is None:
            warn(f'No estimates given. Continuing with default parameter values.')
            _no_ests = True
        else:
            # Check internal dtypes of estimates
            if not Utils.check_dtype(estimates, (float, int)):
                raise ValueError(f'Internals of estimates must be integer or float')
            _no_ests = False

        # Check if input is given
        if estimator_object is None:
            raise IOError('estimator_object must be provided')

        # Check data argument
        if data is None:
            print('Using data of estimator_object...')
            data = estimator_object.data
            rids = estimator_object.rids
        else:
            # Check input structure of data by calling estim8.Estimator.data property
            _temp_estimator         = Estimator(estimator_object.model, estimator_object.bounds)
            _temp_estimator.data    = data
            rids                    = _temp_estimator.rids

        # Checking input datatypes
        if not isinstance(estimates, dict) and not None:
            raise TypeError(f'estimates must be dict-type, not {type(estimates)}')
        if not isinstance(estimator_object,Estimator) and not None:
            raise TypeError(f'estimator_object must be an instance of the estim8.Estimator-class, not {type(estimator_object)}')
        if not isinstance(data,dict) and not isinstance(data,pandas.core.frame.DataFrame) and not None:
            raise TypeError(f'data must be provided as dict or pandas.DataFrame, not {type(data)}')

        # Check matching RIDs
        if estimator_object.rids is not None:
            rid_diff = list(set(data.keys())-set(estimator_object.rids))
            if rid_diff:
                raise ValueError(f'The replicate IDs of the provided dataset do not match those of the estimator_object'
                                 f'\n{rid_diff}\n the rids of the estimator are: {estimator_object.rids}')
        else:
            rids = ['1st']
            data = {'1st':data}


        """ Simulation Step + Replicate Handling """
#TODO: warn if no observation mapping
        # Get time-settings from estimator_object
        t0                      = estimator_object.t[0]
        t_end                   = estimator_object.t[1]
        dt                      = estimator_object.t[2]


        # Retrieve keys of the estimates, and keys of simulation
        est_v                   = [e for e in estimates.values()]
        par_k                   = [str(key) for key in estimator_object.model.parameters.keys()]
        bnd_k                   = [str(bk)  for bk  in estimator_object.bounds.keys()]

        sim                     = {}

        if rids is not None:
            for rid in rids:
                # Sort by mapping and rid
                _par = Utils.replicate_handling(est_v, bnd_k, par_k,
                                                estimator_object.model.parameters, rid,
                                                estimator_object.parameter_mapping)

                # Simulate
                sim[rid], _ = estimator_object.model.simulate(t0, t_end, dt, parameters=_par, for_est=True)

        else:
            # Fill non-estimated values
            _par = Utils.replicate_handling(est_v, bnd_k, par_k,
                                            estimator_object.model.parameters, None,
                                            estimator_object.parameter_mapping)

            # Simulate
            sim, _ = estimator_object.model.simulate(t0, t_end, dt, parameters=_par, for_est=True)


        """ Plotting Step """

        # Define subplot grid-size
        if not only_measured:
            n_state_vars                = len(estimator_object.model.observations)
            _h_size = 3
            _v_size = int(n_state_vars / _h_size) + bool(n_state_vars % _h_size)


            for rid in rids:

                # Setup figure and grid
                _fig, _axs              = plt.subplots(_v_size,_h_size)
                _fig.set_size_inches(18, 4.5 * _v_size)
                _fig.suptitle(rid)

                # Delete unused axes
                if n_state_vars % _h_size != 0:
                    try:
                        [_fig.delaxes(_axs[_v_size - 1, h_pos*(-1)]) for h_pos in range(1,(_h_size-(n_state_vars % _h_size))+1)]
                    except IndexError:
                        [_fig.delaxes(_axs[h_pos*(-1)]) for h_pos in range(1,(_h_size-(n_state_vars % _h_size))+1)]

                # Initialize subplot enumerators
                _h = 0
                _v = 0

                # Cycle colors
                colors = plt.rcParams["axes.prop_cycle"]()

                # Setup plots for all state variables
                for _obs in estimator_object.model.observations:

                    # Select next color
                    c = next(colors)["color"]

                    # Check for available experimental data
                    if _obs in data[rid].keys():
                        _axs[_v,_h].plot(data[rid].index, data[rid][_obs],'ko')
                        #print(data[rid][_obs])
                    if _obs in estimator_object.observation_mapping.values():
                        _dat_name = [k for k in estimator_object.observation_mapping.items() if k[1] == _obs]
                        if _dat_name[0][0] in data[rid].keys():
                            _axs[_v,_h].plot(data[rid].index, data[rid][_dat_name[0][0]], 'ko')
    #                    print(data[rids][estimator_object.observation_mapping[_dat_name]])


                    # Plot simulation data
                    _axs[_v,_h].plot(sim[rid]["Time"], sim[rid][_obs], color=c)
                    _axs[_v,_h].set_xlabel('Time')
                    _axs[_v,_h].set_title(_obs)

                    #Increase grid indices
                    if _h == _h_size - 1:
                        _v, _h = _v + 1, 0  # if _h is at position 1, set zero and increase _v
                    else:
                        _h += 1  # else increase _h

        else:

            # Define size
            n_state_vars = 0
            for rid in data.keys():
                n_state_vars += len(data[rid].keys())
            _h_size = 3
            _v_size = int(n_state_vars / _h_size) + bool(n_state_vars % _h_size)

            # Setup figure and grid
            _added = False
            if _v_size == 1:
                _v_size = 2   # Avoid index error
                _added  = True
            _fig, _axs = plt.subplots(_v_size, _h_size)
            _fig.set_size_inches(18, 4.5 * _v_size)

            # Delete unused axes
            if n_state_vars % _h_size != 0:
                try:
                    if not _added:
                        [_fig.delaxes(_axs[_v_size - 1, h_pos * (-1)]) for h_pos in
                         range(1, (_h_size - (n_state_vars % _h_size)) + 1)]
                    else:
                        [_fig.delaxes(_axs[_v_size - 2, h_pos * (-1)]) for h_pos in
                         range(1, (_h_size - (n_state_vars % _h_size)) + 1)]
                except IndexError:
                    [_fig.delaxes(_axs[h_pos * (-1)]) for h_pos in range(1, (_h_size - (n_state_vars % _h_size)) + 1)]

            # Initialize subplot enumerators
            _h = 0
            _v = 0

            # Cycle colors
            colors = plt.rcParams["axes.prop_cycle"]()
            color_assign = {}

            for rid in rids:
                # Setup plots for all state variables
                for _obs in estimator_object.model.observations:

                    # Init in experimental data check
                    in_exp = False

                    # Select next color
                    c = next(colors)["color"]

                    # Check for available experimental data
                    if _obs in data[rid].keys():
                        _axs[_v, _h].plot(data[rid].index, data[rid][_obs], 'ko')
                        # Plot simulation data
                        if _obs not in color_assign.keys():
                            _axs[_v, _h].plot(sim[rid]["Time"], sim[rid][_obs], color=c)
                            color_assign[_obs] = c
                        else:
                            _axs[_v, _h].plot(sim[rid]["Time"], sim[rid][_obs], color=color_assign[_obs])
                        _axs[_v, _h].set_xlabel('Time')
                        _axs[_v, _h].set_title(str(rid)+':  '+str(_obs))
                        in_exp=True
                    if _obs in estimator_object.observation_mapping.values():
                        _dat_name = [k for k in estimator_object.observation_mapping.items() if k[1] == _obs]
                        if _dat_name[0][0] in data[rid].keys():
                            _axs[_v, _h].plot(data[rid].index, data[rid][_dat_name[0][0]], 'ko')
                            # Plot simulation data
                            if _obs not in color_assign.keys():
                                _axs[_v, _h].plot(sim[rid]["Time"], sim[rid][_obs], color=c)
                                color_assign[_obs] = c
                            else:
                                _axs[_v, _h].plot(sim[rid]["Time"], sim[rid][_obs], color=color_assign[_obs])
                            _axs[_v, _h].set_xlabel('Time')
                            _axs[_v, _h].set_title(str(rid)+':  '+str(_obs))
                            in_exp = True


                    if in_exp:
                        # Increase grid indices
                        if _h == _h_size - 1:
                            _v, _h = _v + 1, 0  # if _h is at position 1, set zero and increase _v
                        else:
                            _h += 1  # else increase _h

            # Delete added row
            if _added:
                [_fig.delaxes(_axs[_v_size - 1, h_pos * (-1)]) for h_pos in range(1,_h_size+1)]

        # Avoid axis overlapping
        plt.subplots_adjust(
            left=0.1,
            right=0.9,
            bottom=0.1,
            top=0.9,
            wspace=0.4,
            hspace=0.4,

        )
        plt.show()

        return sim
        
        
    @staticmethod
    def plot_correlations(mcs, thresholds:int=5, show_vals:bool=False):
        """
        Plotting the correlations for parameter results of a Monte Carlo Sampling as a heatmap. The threshold defines
        the resolution of values, i.e. it defines a color map with discrete increments. The larger the thresholds
        argument is set, the finer is the resolution. Set show_vals True to write the exact correlation values into the
        plot.

        Arguments:
        ----------
            mcs         : pd.DataFrame / list[dict]
                Parameter information from Monte Carlo Sampling

        Keyword Arguments:
        ------------------
            thresholds  : int
                Number of increments for color map
            show_vals   : bool
                For writing exact values into the plot

        External Functions:
        -------------------
            corr_mat    : utils.py
                Computes the calculations between each key in the parameter data, and creates a pivot table of those
        """
        
        # Check input
        if not isinstance(mcs, pd.core.frame.DataFrame):
            try:
                mcs = pd.DataFrame(mcs)
            except:
                raise ValueError(f'data could not be converted into DataFrame. Pleas try passing a DataFrame instead')
            
        # Compute Correlations
        corrs = Utils.corr_mat(mcs)
        
        # Setup figure
        sns.set(rc={'figure.figsize': (22, 22)})
        sns.set(font_scale=2.5)

        # Setup colors
        myColors = []
        th       = np.linspace(-1,1,thresholds)
        for i in range(thresholds):
            if i < (thresholds/2):
                myColors.append(tuple([0.0,0.0,0.8,abs(th[i])]))
            else:
                myColors.append(tuple([0.8,0.0,0.0,abs(th[i])]))

            cmap = LinearSegmentedColormap.from_list('Custom', myColors, len(myColors))


        # Plot Heatmap
        sns.heatmap(
            data=corrs,
            vmin=-1,
            vmax=1,
            center=0,
            square=True,
            annot=show_vals,
            cmap=cmap,
            linecolor='black',
            linewidths=0.1
        )

        plt.show()


    @staticmethod
    def plot_distribution(mcs, bins:int=5, est:Estimator=None):
        """
        Plots the distributions of Monte Carlo Samples as hist-plot for the individual occurrence and as scatter plot
        for pair-wise comparison. If an Estimator object is supplied, the scatter plot is highlighted in blueish colors
        for low losses and reddish colors for high losses.
        TODO: Color scale for Loss
        FIXME: check input in advance
        Arguments:
        ----------
             mcs         : pd.DataFrame / list[dict]
                Parameter information from Monte Carlo Sampling

        Keyword Arguments:
        ------------------
            bins        : int
                Number of bins for hist plot
            est         : Estimator
                Estimator object for loss recalculation
        """

        # Check or convert dtype
        if not isinstance(mcs,pd.core.frame.DataFrame):
            mcs = pd.DataFrame(mcs)

        if est is not None:
            loss = {}
            for ind in mcs.index:
                loss[ind] = est.obj(mcs.loc[ind,:])

        # Make a keylist
        mcs_keys = [str(k) for k in mcs.keys()]

        # Initialize plot framework
        size        = len(mcs.keys())
        fig, axs    = plt.subplots(size,size)
        fig.set_size_inches(4*size, 3*size)
        plt.subplots_adjust(
            left=0.1,
            right=0.9,
            bottom=0.1,
            top=0.95,
            wspace=0.4,
            hspace=0.4,
        )

        # Iterate over length of keys
        for i in range(size):

            # Set Grid titles
            axs[i,i].set_title(mcs_keys[i])
            if i != 0:
                axs[i,0].set_ylabel(mcs_keys[i])

            # Plot histogram
            axs[i,i].hist(mcs[mcs_keys[i]], bins=bins, rwidth=0.8)

        for i in range(size):
            for j in range(size):
                if i < j:
                    axs[i,j].set_visible(False)
                elif i > j:
                    try:
                        c = [float(l) for l in loss.values()]
                        axs[i,j].scatter(x=mcs[mcs_keys[j]], y=mcs[mcs_keys[i]], c=c, cmap='jet', alpha=0.6)
                        axs[i,j].scatter(x=np.median(mcs[mcs_keys[j]]), y=np.median(mcs[mcs_keys[i]]), color='black',
                                          marker='X', linewidths=1.5)
                    except:
                        axs[i,j].scatter(x=mcs[mcs_keys[j]], y=mcs[mcs_keys[i]])
                        axs[i,j].scatter(x=np.median(mcs[mcs_keys[j]]), y=np.median(mcs[mcs_keys[i]]), color='red',
                                          marker='X', linewidths=1.5)


        plt.show()


    @staticmethod
    def plot_many(mcs, est:Estimator, observe:list=None):
        """
        Plots the simulation results for all Monte Carlo Samples for the observations of interest.
        """

        # Check data
        if not isinstance(mcs, pd.core.frame.DataFrame):
            mcs = pd.DataFrame(mcs)
        if observe is None:
            observe = est.model.observations

        # RECOMPUTE SIMULATION RESULTS
        sim = []
        for ind in mcs.index:
            mcs_sim = {}
            if est.rids is not None:
                for rid in est.rids:
                    # Treat replicates
                    if est.parameter_mapping is not None:
                        # Apply mapping FIXME: Use iloc to avoid problems with double indices
                        par = Utils.replicate_handling(mcs.loc[ind,:].values, list(est.bounds.keys()), est.model.parameters.keys(),
                                                       est.model.parameters, rid, est.parameter_mapping)
                    else:
                        # Use Sampling data directly
                        par = mcs.loc[ind,:].to_dict()

                    # Simulate
                    mcs_sim[rid], _ = est.model.simulate(est.t[0],est.t[1],est.t[2], parameters=par)
            else:
                mcs_sim['1st'], _   = est.model.simulate(est.t[0],est.t[1],est.t[2], parameters=mcs.loc[ind,:].to_dict())

            # Save simulation results as sim[mcs][rid][quant.]
            sim.append(mcs_sim)


        # READ EXPERIMENTAL DATA

        # Convert to dict if necessary
        if not isinstance(est.data, dict):
            _exp_dat = {'1st':est.data}
        else:
            _exp_dat = est.data

        # Extract observation mapping
        obs_map  = est.LC.observation_mapping

        # Check if observation mapping is specified and apply
        if obs_map is not None:
            exp_dat = {}
            for rid in _exp_dat.keys():
                dat = {}
                for q in _exp_dat[rid].keys():
                    dat[obs_map[q]] = _exp_dat[rid][q]
                exp_dat[rid] = dat
        else:
            exp_dat = _exp_dat

        # SETUP PLOTS

        # Size pre-settings
        n_plots = len(observe)                      # Number of plots per replicate
        n_row   = int(n_plots/2) + (n_plots%2!=0)   # Required number of rows
        if len(sim) < 50:                           # Set line width according to sample number
            wd  = 0.4
        else:
            wd  = 0.2

        # Iterate over replicates
        for rid in sim[0].keys():

            # Create subplots and set size
            fig, axs = plt.subplots(n_row, 2)
            fig.set_size_inches(18, 4 * n_row)
            fig.suptitle(rid, size='xx-large')

            # Check if axis is big enough
            if n_row > 1:

                # Initialize counters
                i, j = 0, 0

                # Iterate over observations
                for obs in observe:
                    # Plot all samples in red
                    [axs[i,j].plot(sim[s][rid]["Time"], sim[s][rid][obs], color='red', linewidth=wd) for s in range(len(sim))]
                    axs[i,j].set_ylabel(obs)
                    axs[i, j].set_xlabel("Time")

                    # Plot experimental data
                    if obs in exp_dat[rid].keys():
                        axs[i,j].plot(exp_dat[rid][obs], 'bx')

                    if j == 0:
                        i, j = i, j + 1
                    else:
                        i, j = i + 1, 0
            else:
                i = 0
                for obs in observe:
                    # Plot all samples in red
                    [axs[i].plot(sim[s][rid]["Time"], sim[s][rid][obs], color='red', linewidth=wd) for s in range(len(sim))]
                    axs[i].set_ylabel(obs)
                    axs[i].set_xlabel("Time")

                    # Plot experimental data
                    if obs in exp_dat[rid].keys():
                        axs[i].plot(exp_dat[rid][obs], 'bx')

                    i += 1


        plt.subplots_adjust(
            left=0.1,
            right=0.9,
            bottom=0.1,
            top=0.95,
            wspace=0.4,
            hspace=0.4,
        )

        plt.show()

    @staticmethod
    def plot_bound_violation(est, bounds):


        Y_percent = {}
        Y_ub = {}
        Y_lb = {}
        N_ub = {}
        N_lb = {}
        ind = []

        for key in est.keys():
            Y_percent[key] = ((est[key] - bounds[key][0]) / (
                        bounds[key][1] - bounds[key][0]) * 100 - 50) * 2
            Y_ub[key] = 100
            Y_lb[key] = -100
            N_ub[key] = ((bounds[key][1] - bounds[key][0]) / (
                        bounds[key][1] - bounds[key][0]) * 100 - 50) * 2
            N_lb[key] = ((bounds[key][0] - bounds[key][0]) / (
                        bounds[key][1] - bounds[key][0]) * 100 - 50) * 2
            ind.append(str(key))

        plt.figure(figsize=(20, 10), dpi=80)

        p1 = plt.scatter(Y_percent.keys(), Y_percent.values())
        p2 = plt.plot(Y_ub.keys(), Y_ub.values(), 'b-.')
        p3 = plt.plot(Y_lb.keys(), Y_lb.values(), 'b-.')
        # p4 = plt.bar(.keys(),list(N_ub.values()), width=0.2, bottom=list(N_lb.values()), color='g')
        p5 = plt.plot(N_ub.keys(), N_ub.values(), 'g-')
        p6 = plt.plot(N_lb.keys(), N_lb.values(), 'g-')

        plt.title('Parameter Bounds', fontsize=24)
        plt.xticks(ind, rotation=90, fontsize=18)
        plt.yticks([-200, -175, -150, -125, -100, -75, -50, -25, 0, 25, 50, 75, 100, 125, 150, 175, 200], fontsize=18)
        plt.show()
